# Copyright 2011 Canonical Ltd
#
from ConfigParser import ConfigParser
from argparse import ArgumentParser
from launchpadlib.launchpad import Launchpad
from lp2kanban.cards2workitems import (
    convert_cards_to_work_items,
    get_blueprint_linked_cards,
    update_blueprints_from_work_items,
    )
from lp2kanban.kanban import (
    LeankitKanban,
    Record,
    )
import os.path
import re
import sys
from traceback import print_exc

TITLE_MARKER = "sync:"
DESCRIPTION_MARKER = "Synced with Launchpad"
NO_SYNC_MARKER = "(no-sync)"


def get_arg_parser():
    parser = ArgumentParser(
        description="Update a Kanban board from Launchpad data.")
    parser.add_argument(
        '-c', '--config', default='config.ini',
        help="LeankitKanban configuration file.")
    parser.add_argument(
        '-e', '--email', required=False, default=None,
        help="The Kanban user's registered email address")
    parser.add_argument(
        '-p', '--password', required=False, default=None,
        help="The user's Kanban password")
    parser.add_argument(
        '-b', '--board', required=False,
        help="The Kanban board (a title).")
    parser.add_argument(
        '-l', '--list', action='store_const', const=True, default=False,
        help="List available kanban boards.")
    parser.add_argument(
        '-d', '--debug', action='store_const', const=True, default=False,
        help="Turn on debugging for HTTP traffic")
    return parser


def parse_config(config):
    """Parse the kanban configuration from `config`.

    GLOBAL section is used for global settings for accessing the kanban
    board.  DEFAULT contains the default values which are included in
    each of the board sections automatically.

    Other sections should be named after the board title that they
    are configuring.
    """
    globals = {}
    for key, value in config.items('GLOBAL'):
        globals[key] = value

    boards = {}
    for section in config.sections():
        if section in ['GLOBAL']:
            continue
        boards[section] = {}
        for key in config.defaults():
            boards[section][key] = config.get(section, key)
        for key in config.options(section):
            boards[section][key] = config.get(section, key)
    return globals, boards


def parse_groups_config(config):
    """Parse the groups configuration file from `config`.

    The groups config is an ini file, where the sections are the group
    names, and the `members` attribute of each group is a comma
    separated list of Launchpad names.

    Returns a dict mapping a Launchpad name to a group name.
    """
    lp_to_group = {}
    for group in config.sections():
        if not config.has_option(group, "members"):
            continue
        for member in config.get(group, "members").split(","):
            lp_to_group[member.strip()] = group
    return lp_to_group


class LaunchpadUsersForBoard(object):
    """Find Launchpad user accounts for board users and keep them cached.

    Provides three dicts:
     * lp_to_kanban which uses Launchpad account names as keys and points
       to LeankitUser records.
     * kanban_to_lp which uses LeankitUser records as keys pointing to
       actual Launchpad user records.
     * lp_to_group which uses Launchpad account names as keys and points
       to group names.
    """
    def __init__(self, kanban_to_lp=None, lp_to_kanban=None, lp_to_group=None):
        if kanban_to_lp is None:
            kanban_to_lp = {}
        self.kanban_to_lp = kanban_to_lp
        if lp_to_kanban is None:
            lp_to_kanban = {}
        self.lp_to_kanban = lp_to_kanban
        if lp_to_group is None:
            lp_to_group = {}
        self.lp_to_group = lp_to_group

    def set_up_users(self, lp, board, config):
        """Set up the dicts this object provides."""
        for user in board.users:
            lp_user = lp.people.getByEmail(email=user)
            if lp_user is None:
                print (
                    "WARNING: There is no matching Launchpad user for kanban"
                    " user %s." % (user,))
                continue
            self.lp_to_kanban[lp_user.name] = board.users[user]
            self.kanban_to_lp[board.users[user]] = lp_user
        groups_config_file = config.get("groups_config_file")
        if groups_config_file:
            assert os.path.exists(groups_config_file), (
                "Groups config file not found: " + groups_config_file)
            groups_config = ConfigParser()
            groups_config.read(groups_config_file)
            self.lp_to_group.update(parse_groups_config(groups_config))


ORDERED_STATUSES = {'Rejected': 0,
                    'Work in progress': 0,
                    'Needs review': 1,
                    'Approved': 2,
                    'Merged': 2,
                    }
REVIEW_STATUS = 'In Review'
TRANSFORM_STATUS = {'Rejected': 'In Progress',
                    'Work in progress': 'In Progress',
                    'Needs review': REVIEW_STATUS,
                    'Approved': 'Approved',
                    'Merged': 'Merged',
                    }
FINAL_STATUSES = ('Approved', 'Merged')


def get_branch_info(branches):
    """Returns a bug's status based on the given branches.

    Result will be one of "Approved", "Merged", "In Review", "In
    Progress", or None.  The meanings of the result values should be
    obvious with the addition that None signifies that no information
    could be gleaned from the branches.
    """
    info = Record(status=None, target=None, mp_url=None, mp_type=None)
    if len(branches) > 1:
        # We don't know what to do if the bug has more than one branch, so
        # refuse to guess.
        return info

    def _get_mp_info():
        mps = []
        for bug_branch in branches:
            for mp in bug_branch.branch.landing_targets:
                mp_info = Record(rank=None, status=None, mp=None)
                status = mp.queue_status
                mp_info.rank = ORDERED_STATUSES.get(status, 1)
                mp_info.status = TRANSFORM_STATUS.get(status, REVIEW_STATUS)
                mp_info.mp = mp
                # If a merge proposal is in a final state, use it and stop
                # looking.
                if status in FINAL_STATUSES:
                    return mp_info
                mps.append(mp_info)
        if len(mps) == 0:
            return None
        mps.sort(key=lambda x:(x.rank, x.mp.date_created), reverse=True)
        return mps[0]

    def _get_mp_link(mp):
        regex = re.compile('(https://codereview.appspot.com/\d+)')
        match = regex.search(mp.description or '')
        if match:
            # An appspot Rietveld URL was found.  Use it.
            url = match.group(0)
            mp_type = 'Rietveld'
        else:
            url = mp.web_link
            mp_type = 'LP Merge Proposal'
        return (url, mp_type)

    mp_info = _get_mp_info()
    if mp_info is None:
        return info
    mp = mp_info.mp
    info.status = mp_info.status
    info.target = mp.target_branch.name
    info.mp_url, info.mp_type = _get_mp_link(mp)
    return info


def should_sync_card(card, conf={}):
    """Check to see if the card should be synced.

    `conf` is the kanban board configuration as a dict, and should contain
    an option `autosync` set to 'on' if all cards with external card IDs
    should be synced unless their description or title contain NO_SYNC_MARKER.
    With autosync off, if the card title starts with TITLE_MARKER or the card
    description contains DESCRIPTION_MARKER then return True.

    If `conf` has the option `sync_cards` set the 'off', False is returned.
    """
    if conf.get('sync_cards', 'off') == 'off':
        return False
    no_sync_lanes = [
        lane_path.strip()
        for lane_path in conf.get("no_sync_lanes", "").split(",")]
    for no_sync_lane_path in no_sync_lanes:
        if no_sync_lane_path and card.lane.path.endswith(no_sync_lane_path):
            return False
    if conf.get('autosync', None) == 'on':
        has_id = (card.external_card_id is not None and
                  card.external_card_id.strip() != '')
        return (has_id and NO_SYNC_MARKER not in card.title and
                NO_SYNC_MARKER not in card.description)
    else:
        return (card.title.startswith(TITLE_MARKER) or
                DESCRIPTION_MARKER in card.description)


def should_move_card(card, conf={}):
    """Check to see if a card is eligible to be moved.

    The configuration parameter `move_cards` must be enabled and the card must
    be in an active portion of the board.
    """
    enabled = conf.get('move_cards', 'off') == 'on'
    active = not card.lane.board.is_archived
    return enabled and active


NOOP_BUG_STATUSES = [
    "New", "Incomplete", "Confirmed", "Triaged",
]
IN_PROGRESS_BUG_STATUSES = [
    "In Progress", "Fix Committed",
    ]
DONE_BUG_STATUSES = [
    "Fix Released", "Opinion", "Invalid", "Won't Fix", "Expired",
]
ORDERED_BUG_STATUSES = (
    NOOP_BUG_STATUSES + IN_PROGRESS_BUG_STATUSES + DONE_BUG_STATUSES)


class CardStatus:
    """Card is consider to be in only one of these statuses."""
    CODING = 0
    REVIEW = 1
    LANDING = 2
    QA = 3
    DEPLOY = 4
    DOWNTIME_DEPLOY = 5
    DONE = 6


# Configuration variables for individual card statuses.
ConfigOptionByStatus = {
    CardStatus.CODING: 'coding_lanes',
    CardStatus.REVIEW: 'review_lanes',
    CardStatus.LANDING: 'landing_lanes',
    CardStatus.QA: 'qa_lanes',
    CardStatus.DEPLOY: 'deploy_lanes',
    CardStatus.DOWNTIME_DEPLOY: 'downtime_deploy_lanes',
    CardStatus.DONE: 'done_lanes',
    }

def get_new_bugs(board, bugs):
    """Return the bugs that aren't already in the board."""
    existing_bugs = set()
    for card in board.cards:
        if card.external_card_id:
            try:
                bugid = int(card.external_card_id)
            except ValueError:
                print "Invalid external id (%s) for card '%s'" % (
                    card.external_card_id, card.title)
                continue
            existing_bugs.add(bugid)
    for bug in bugs:
        if bug.id not in existing_bugs:
            yield bug


def get_card_status(bug_status, bug_tags, branch_info):
    """Return the status of the card as one of CardStatus values."""
    # status of None means card needs not to be moved.
    status = None
    if bug_status in IN_PROGRESS_BUG_STATUSES:
        status = CardStatus.CODING
        if branch_info.status == 'In Progress':
            status = CardStatus.CODING
        elif branch_info.status == 'In Review':
            status = CardStatus.REVIEW
        elif branch_info.status == 'Approved':
            status = CardStatus.LANDING
        elif branch_info.status == 'Merged':
            if bug_status == 'Fix Committed':
                if 'qa-needstesting' in bug_tags:
                    status = CardStatus.QA
                else:
                    if branch_info.target == 'db-devel':
                        status = CardStatus.DOWNTIME_DEPLOY
                    else:
                        status = CardStatus.DEPLOY
            else:
                status = CardStatus.LANDING
    elif bug_status in DONE_BUG_STATUSES:
        status = CardStatus.DONE
    return status


def get_bug_status(lp_bug, all_projects, all_users):
    lowest_status = None
    matches = False
    assignees = []
    tasks = lp_bug.bug_tasks_collection
    for task in tasks:
        if (task.target in all_projects or
            (task.assignee is not None and
             task.assignee.name in all_users)):
            matches = True
            if (lowest_status is None or
                (ORDERED_BUG_STATUSES.index(task.status) <
                 ORDERED_BUG_STATUSES.index(lowest_status))):
                lowest_status = task.status
            if task.assignee is not None:
                assignees.append(task.assignee)
    branch_info = get_branch_info(lp_bug.linked_branches)
    status = get_card_status(lowest_status, lp_bug.tags, branch_info)
    return matches, status, assignees, branch_info.mp_url, branch_info.mp_type


def find_card_target_lane_omnidirectional(lane, path):
    """Find the target lane for the path using the "omnidirectional" algorithm.

    The "omnidirectional" algorithm searches for the target lane from
    the beginning of the board, no matter what the current lane is.
    """
    for possible_target_lane in lane.board.lanes.values():
        if possible_target_lane.path == path:
            return possible_target_lane



def find_card_target_lane_next(lane, path):
    """Find the target lane for the path using the "next" algorithm.

    The "next" algorithm searches only the current lane's next lanes.
    """
    if lane.path.endswith(path):
        return lane
    else:
        next_lanes = lane.getNextLanes()
        if next_lanes is None:
            return None

        for next_lane in next_lanes:
            target_lane = find_card_target_lane_next(next_lane, path)
            if target_lane is not None:
                return target_lane
        return None

def create_new_cards(board, launchpad_project, bug_tag, cardtype_name=None,
                     cardlane_path=None):
    """Create cards for bugs tagged in Launchpad.

    If a bug is tagged with the given tag, a new card will be created in
    the board and the tag will be removed from the bug.
    """
    new_cardtype = board.default_cardtype.id
    if cardtype_name is not None:
        for type_id, cardtype in board.cardtypes.items():
            if cardtype.name == cardtype_name:
                new_cardtype = type_id
    kanban_bugtasks = launchpad_project.searchTasks(tags=[bug_tag])
    kanban_bugs = get_new_bugs(
        board, [bugtask.bug for bugtask in kanban_bugtasks])
    if cardlane_path is not None:
        next_lane = board.getLaneByPath(cardlane_path)
    else:
        next_lane = board.root_lane
        while next_lane.child_lanes:
            next_lane = next_lane.child_lanes[0]
    for bug in kanban_bugs:
        print "    Creating card for bug %s in %s." % (bug.id, next_lane.title)
        card = next_lane.addCard()
        card.title = bug.title
        card.description = bug.description
        card.external_card_id = str(bug.id)
        card.type_id = new_cardtype
        card.save()
        bug.tags = [tag for tag in bug.tags if tag != bug_tag]
        bug.lp_save()


def get_lp(config):
    """Get a lp instance for a logged in user or anonymous access.

    If 'credentials_file' is specified in the board configuration then an
    instance of Launchpad for the user is returned, otherwise anonymous access
    is used.
    """
    credentials_file = config.get('credentials_file')
    launchpadlib_dir = config.get('launchpadlib_dir')
    if credentials_file and os.path.exists(credentials_file):
        lp = Launchpad.login_with(
            'lp2kanban', config['lp_server'], version="devel",
            credentials_file=credentials_file,
            launchpadlib_dir=launchpadlib_dir)
        print "   Logged in as", lp.me.name
    else:
        lp = Launchpad.login_anonymously(
            'lp2kanban', config['lp_server'], version="devel",
            launchpadlib_dir=launchpadlib_dir)
        print ("   Accessing Launchpad anonymously.  "
               "Private bugs will not be synced.")
    return lp


def sync_board(board, bconf):
    lp = get_lp(bconf)
    all_projects = []
    project_group = bconf.get('project_group')
    if project_group:
        launchpad_project_group = lp.project_groups[project_group]
        # These are sometimes very slow (and basically take a considerable
        # chunk of the runtime).  We should optimize this so it's done only
        # once per run instead of once per board being synced.
        all_projects = list(launchpad_project_group.projects_collection)
    project_names = bconf.get('projects')
    if project_names:
        for name in project_names.split(','):
            project = lp.projects[name.strip()]
            if project:
                all_projects.append(project)
    lp_users = LaunchpadUsersForBoard()
    lp_users.set_up_users(lp, board, bconf)
    new_bug_tag = bconf.get("bug_to_card_tag")
    if new_bug_tag is not None:
        for project in all_projects:
            print "   Creating new cards for %s:" % project.name
            create_new_cards(
                board, project, new_bug_tag, bconf.get("bug_to_card_type"),
                bconf.get("bug_to_card_lane"))
    print "   Syncing cards:"
    for card in board.getCardsWithExternalIds():
        if should_sync_card(card, bconf):
            try:
                bug_id = int(card.external_card_id)
            except ValueError:
                print "Invalid external id (%s) for card '%s'" % (
                    card.external_card_id, card.title)
                continue
            try:
                lp_bug = lp.bugs[bug_id]
            except KeyError:
                print (
                    "   * %d is not accessible "
                    "(does not exist or forbidden)." % (bug_id))
                continue
            synced, card_status, assignees, mp_url, mp_type = get_bug_status(
                lp_bug, all_projects, lp_users.lp_to_kanban.keys())

            # Move card to new lane if configured to do so.
            if card_status is not None and should_move_card(card, bconf):
                try:
                    move_card(card, card_status, bconf, assignees, lp_users)
                except IOError as e:
                    print "   * %s: %s (in %s)" % (
                        bug_id, card.title, card.lane.path)
                    print "     >>> Error moving card:"
                    print "     ", e
                    continue

            if synced:
                card.title = lp_bug.title
                for tag in lp_bug.tags:
                    card.addTag(tag)
                if len(assignees) > 0:
                    assignee = assignees[0]
                    kanban_user = lp_users.lp_to_kanban.get(
                        assignee.name, None)
                    if kanban_user:
                        card.assigned_user_id = kanban_user.id
                        card.kanban_user = kanban_user
                if mp_url and mp_type:
                    card.external_system_url = mp_url
                    card.external_system_name = mp_type
                print "   * %s: %s (in %s)" % (
                    bug_id, card.title, card.lane.path)
            else:
                print "   * %s: %s -- Not synced" % (
                    bug_id, card.title)
    if bconf.get("convert_cards_to_work_items", False) == "on":
        # Sync cards to blueprint work items, if so configured.
        print "   Checking for cards to sync with work items..."
        cards_to_convert = get_blueprint_linked_cards(board)
        work_items = convert_cards_to_work_items(
            cards_to_convert, board, lp_users)
        if len(work_items) == 0:
            print "No cards converted to work items."
        else:
            print "Updating blueprints from work items."
            update_blueprints_from_work_items(lp, work_items)


def move_card(card, bug_status, bconf, assignees, lp_users):
    """Move a card to its right place on the board.

    The right place is considered by looking at the status and the first
    assignee. If the `omnidirectional_card_moves` is turned on in the board
    config, the card can be moved to any lane on the board. If it's not
    turned on, only the lanes that are considered to be "next" of the
    card's current lane are possible move targets.

    If a card is located in a "no_move" lane, the card won't get moved.
    """
    target_lane_path = bconf[ConfigOptionByStatus[bug_status]]
    if assignees:
        group = lp_users.lp_to_group.get(assignees[0].name)
        if group:
            target_lane_path = target_lane_path.replace("${group}", group)
    omnidirectional_card_moves = (
        bconf.get("omnidirectional_card_moves", "on") == "on")
    no_move_lanes = [
        lane_path.strip()
        for lane_path in bconf.get("no_move_lanes", "").split(",")]
    for no_move_lane_path in no_move_lanes:
        if no_move_lane_path and card.lane.path.endswith(no_move_lane_path):
            return
    if omnidirectional_card_moves:
        target_lane = find_card_target_lane_omnidirectional(
            card.lane, target_lane_path)
    else:
        target_lane = find_card_target_lane_next(
            card.lane, target_lane_path)
    card.move(target_lane)


def save_board(board):
    # Split out so we can change the call to a single UpdateCards call.
    for card in board.getCardsWithExternalIds():
        if card.is_dirty:
            if DESCRIPTION_MARKER not in card.description:
                card.description += '\n' + DESCRIPTION_MARKER
                card.description = card.description.strip()

            print "     Saving '%s' for #%s." % (
                "', '".join(card.dirty_attrs), card.external_card_id)
        # Attempt to save the card.  If is it not dirty or new then nothing
        # happens.
        try:
            card.save()
        except IOError as e:
            print "Exception caught.  Skipping board."
            print "  desc:", card.description
            print "  kanban id:", card.assigned_user_id
            print "  LP bug id:", card.external_card_id
            print_exc()
            continue


def main(argv):
    """Update the Yellow team's kanban board with data from Launchpad."""
    args = get_arg_parser().parse_args()

    if args.debug:
        import httplib
        httplib.HTTPConnection.debuglevel = 1

    config = ConfigParser()
    config.read(args.config)
    global_config, boards_config = parse_config(config)
    user = args.email or global_config['user']
    password = args.password or global_config['password']
    kanban = LeankitKanban(global_config['account'], user, password)

    if args.list:
        print "All active boards:"
        boards = kanban.getBoards()
        for board in boards:
            print " * %s (%d)" % (board.title, board.id)
        return

    if args.board:
        assert args.board in boards_config, (
            "Requested board '%s' is not present in the config file %s." % (
                args.board, args.config))
        board_names = [args.board]
    else:
        board_names = boards_config.keys()
    print "Requested boards:"
    for board_name in board_names:
        print " * %s" % board_name
        try:
            board = kanban.getBoard(title=board_name)
            if board is None:
                print "Could not retrieve board.  Skipping."
                continue
            sync_board(board, boards_config[board_name])
            save_board(board)
        except (KeyError, AssertionError) as e:
            print "Exception caught.  Skipping board."
            print_exc()
            continue


if __name__ == '__main__':
    main(sys.argv)
