# Copyright 2011-2012 Canonical Ltd
"""Tests for the bugs2cards.py script."""

__metaclass__ = type

from ConfigParser import (
    ConfigParser,
    NoSectionError,
    )
from StringIO import StringIO
import tempfile
from textwrap import dedent
import datetime
import unittest

from lp2kanban.bugs2cards import (
    CardStatus,
    create_new_cards,
    find_card_target_lane_next,
    get_branch_info,
    get_bug_status,
    get_card_status,
    get_new_bugs,
    LaunchpadUsersForBoard,
    move_card,
    parse_config,
    parse_groups_config,
    should_sync_card,
    should_move_card,
    )
from lp2kanban.kanban import (
    LeankitCard,
    Record,
    )
from lp2kanban.tests.common import (
    FauxBoard,
    FauxCard,
    FauxCardType,
    FauxLane,
    )



def parse_config_text(text):
    config = ConfigParser()
    config.readfp(StringIO(dedent(text)))
    return config


class ConfigReadingTest(unittest.TestCase):

    def test_parse_config_no_globals(self):
        # When 'GLOBAL' section is missing from the configuration,
        # an exception is raised.
        sample_conf = '''
            [Test]
            var = board value
        '''

        config = parse_config_text(sample_conf)

        # NoSectionError is raised because there is no obligatory GLOBAL
        # configuration section.
        self.assertRaises(NoSectionError, parse_config, config)

    def test_parse_config_globals(self):
        # Values in the 'GLOBAL' section are returned separately and
        # are independent from per-board configuration.
        sample_conf = '''
            [GLOBAL]
            var1 = value1
            var2 = value2
            [Board 1]
            var1 = board value 1
        '''

        config = parse_config_text(sample_conf)

        globals, boards = parse_config(config)

        self.assertEqual({'var1': 'value1',
                          'var2': 'value2',
                           }, globals)
        self.assertEqual({'Board 1': {'var1': 'board value 1'}}, boards)

    def test_parse_config_defaults(self):
        # Values from the DEFAULT section are always added to every
        # board configuration section as well.
        sample_conf = '''
            [GLOBAL]
            [DEFAULT]
            var = board value
            [Board 1]
        '''

        config = parse_config_text(sample_conf)

        globals, boards = parse_config(config)
        self.assertEqual({'Board 1': {'var': 'board value'}}, boards)

    def test_parse_config_defaults_overridden(self):
        # Each board can override settings from the defaults.
        sample_conf = '''
            [GLOBAL]
            [DEFAULT]
            var = board value
            [Board 1]
            var = my board value
        '''

        config = parse_config_text(sample_conf)

        globals, boards = parse_config(config)
        self.assertEqual({'Board 1': {'var': 'my board value'}}, boards)


class FauxBugBranch:
    def __init__(self, mps):
        self.branch = Record(landing_targets=mps)


class FauxMP:
    def __init__(self, queue_status, target_name, link=None, description=None, date_created=None):
        self.queue_status = queue_status
        self.target_branch = Record(name=target_name)
        self.web_link = link
        self.description = description or ''
        self.date_created = date_created


class BranchInfoTests(unittest.TestCase):

    def test_no_guessing_on_multiple_branches(self):
        # If a bug has more than one branch associated with it we act as if it
        # has no branches because the code doesn't (yet) know how to interpret
        # more than one branch.
        result = get_branch_info([None, None])
        self.assertEqual(result, {'status': None, 'target': None,
                                  'mp_url': None, 'mp_type': None})

    def test_no_branches_mean_no_info(self):
        # If there are no branches associated with a bug, then no info is
        # returned.
        result = get_branch_info([])
        self.assertEqual(result, {'status': None, 'target': None,
                                  'mp_url': None, 'mp_type': None})

    def test_branches_returns_target(self):
        # The MP's target is pulled out of the bug branch and returned by
        # get_branch_info.
        branches = [FauxBugBranch([FauxMP('Approved', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['target'], 'devel')

    def test_branches_returns_mp_link(self):
        # The MP's link is returned as the mp_url.
        branches = [FauxBugBranch([FauxMP('Approved', 'devel',
                                          'http://example.com/mp/link')])]
        result = get_branch_info(branches)
        self.assertEqual(result['mp_url'], 'http://example.com/mp/link')

    def test_branches_returns_rietveld_link(self):
        # If, however, a link to https://coderview.appspot.com/\d+ is found
        # in the MP description, then that URL is returned.
        branches = [FauxBugBranch([FauxMP(
            'Approved', 'devel', 'http://example.com/mp/link',
            'https://codereview.appspot.com/1234')])]
        result = get_branch_info(branches)
        self.assertEqual(result['mp_url'],
                         'https://codereview.appspot.com/1234')

    def test_branches_in_review(self):
        # If there is an MP then normally the bug's status is considered "In
        # Review".
        branches = [FauxBugBranch([FauxMP('Fake', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'In Review')

    def test_approved_branches_are_approved(self):
        branches = [FauxBugBranch([FauxMP('Approved', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'Approved')

    def test_merged_branches_are_merged(self):
        branches = [FauxBugBranch([FauxMP('Merged', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'Merged')

    def test_in_progress_branches_are_in_progress(self):
        branches = [FauxBugBranch([FauxMP('Work in progress', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'In Progress')

    def test_rejected_branches_are_in_progress(self):
        branches = [FauxBugBranch([FauxMP('Rejected', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'In Progress')

    def test_highest_ranked_returned(self):
        # If there are multiple merge proposals, the one with the highest rank
        # wins.
        branches = [FauxBugBranch([FauxMP('In Review', 'devel'),
                                   FauxMP('Approved', 'approved-branch'),
                                   FauxMP('Rejected', 'bad-branch'),
                                   ])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'Approved')
        self.assertEqual(result['target'], 'approved-branch')

    def test_first_highest_ranked_returned(self):
        # If there are multiple merge proposals of the same highest rank, the
        # newest one wins.
        branches = [FauxBugBranch(
            [FauxMP('In Review', 'devel'),
             FauxMP('Approved', 'first'),
             FauxMP('Approved', 'second'),
             FauxMP('Rejected', 'bad-branch'),
             ])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'Approved')
        self.assertEqual(result['target'], 'first')

    def test_newest_highest_ranked_returned(self):
        # If there are multiple merge proposals of the same highest rank, but
        # not in final states, the newest one wins.
        now = datetime.datetime.now()
        yesterday = now - datetime.timedelta(days=1)
        lastweek = now - datetime.timedelta(days=7)
        branches = [FauxBugBranch(
            [FauxMP('Needs review', 'lastweek', date_created=lastweek),
             FauxMP('Needs review', 'now', date_created=now),
             FauxMP('Needs review', 'yesterday', date_created=yesterday),
             FauxMP('Rejected', 'bad-branch'),
             ])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'In Review')
        self.assertEqual(result['target'], 'now')

    def test_first_seen_final_state_returned(self):
        # If there are multiple MPs in a final state, the first one seen wins.
        branches = [FauxBugBranch([FauxMP('In Review', 'devel'),
                                   FauxMP('Merged', 'merged-branch'),
                                   FauxMP('Approved', 'approved-branch'),
                                   ])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'Merged')


class FauxBug:
    def __init__(self, tags=[], title="Faux Bug",
                 description="This is a faux bug.",
                 bug_id=None):
        self.bug_tasks_collection = []
        self.linked_branches = []
        self.tags = tags
        self.title = title
        self.description = description
        self.id = bug_id
        self.saved = False

    def addTask(self, target, status='New', assignee=None):
        record = Record(target=target, status=status, assignee=assignee)
        self.bug_tasks_collection.append(record)
        return record

    def lp_save(self):
        self.saved = True


class FauxBugTask:

    def __init__(self, bug_id, target='faux-project', tags=None,
                 title="Faux Bug", description="This is a faux bug."):
        if tags is None:
            tags = []
        self.bug = FauxBug(
            bug_id=bug_id, title=title, description=description, tags=tags)
        self.record = self.bug.addTask(target)


class BugStatusTest(unittest.TestCase):

    def test_get_bug_status_unknown(self):
        # For a bug on an unrelated project and not assigned to any
        # of the relevant users, it is not considered to match at all.
        bug = FauxBug()
        bug.addTask('project')
        matches, status, assignees, mp_url, mp_type = (
            get_bug_status(bug, ['launchpad'], []))
        self.assertFalse(matches)
        self.assertEqual(None, status)
        self.assertEqual([], assignees)

    def test_get_bug_status_assignee(self):
        # For a bug on an unrelated project and assigned to a user
        # we care about, it is returned as matching with proper status.
        bug = FauxBug()
        user = Record(name='user')
        bug.addTask('project', status='In Progress', assignee=user)
        matches, status, assignees, mp_url, mp_type = (
            get_bug_status(bug, ['launchpad'], ['user']))
        self.assertTrue(matches)
        self.assertEqual(CardStatus.CODING, status)
        self.assertEqual([user], assignees)

    def test_get_bug_status_project(self):
        # For a bug on an project we care about, it is returned
        # as matching and with proper status.
        bug = FauxBug()
        nonuser = Record(name='nonuser')
        bug.addTask('launchpad', status='In Progress', assignee=nonuser)
        matches, status, assignees, mp_url, mp_type = (
            get_bug_status(bug, ['launchpad'], []))
        self.assertTrue(matches)
        self.assertEqual(CardStatus.CODING, status)
        self.assertEqual([nonuser], assignees)

    def test_get_bug_status_multi_targets(self):
        # For a bug with multiple tasks, only some of which are on
        # projects we care about, only assignees from those tasks are
        # returned.
        bug = FauxBug()
        nonuser = Record(name='nonuser')
        user = Record(name='user')
        bug.addTask('project', status='In Progress', assignee=nonuser)
        bug.addTask('launchpad', status='Fix Released', assignee=user)
        matches, status, assignees, mp_url, mp_type = (
            get_bug_status(bug, ['launchpad'], []))
        self.assertTrue(matches)
        self.assertEqual(CardStatus.DONE, status)
        self.assertEqual([user], assignees)

    def test_get_bug_status_multi_statuses(self):
        # For a bug with multiple tasks, all of which are on
        # projects we care about, the bug status is considered to be the
        # "lowest" of all bug task statuses (i.e. 'In Progress' is lower
        # than 'Fix Committed').
        bug = FauxBug()
        bug.addTask('lazr', status='Fix Released')
        bug.addTask('launchpad', status='In Progress')
        matches, status, assignees, mp_url, mp_type = get_bug_status(
            bug, ['launchpad', 'lazr'], [])
        self.assertTrue(matches)
        # 'Fix Released' on 'lazr' would put the bug status in DONE state,
        # but since there is an 'In Progress' task as well, it is instead
        # in CODING.
        self.assertEqual(CardStatus.CODING, status)


class CardStatusTest(unittest.TestCase):

    def test_should_sync_card_no(self):
        # Cards with sync markers in neither title nor description are
        # not synced.
        card = Record(title=u'no sync', description=u'really no sync')
        self.assertFalse(should_sync_card(card, {'sync_cards': 'on'}))

    def test_should_sync_card_title_yes(self):
        # Sync marker for the title is a string 'sync:' at the beginning
        # of the title string.
        card = Record(title=u'sync:', description=u'description')
        self.assertTrue(should_sync_card(card, {'sync_cards': 'on'}))

    def test_should_sync_card_title_no(self):
        # Sync marker for the title is a string 'sync:' at the beginning
        # of the title string, anywhere else and it's ignored.
        card = Record(title=u'no sync:', description=u'description')
        self.assertFalse(should_sync_card(card, {'sync_cards': 'on'}))

    def test_should_sync_card_description_yes(self):
        # Sync marker for the description is a string 'sync:' at the beginning
        # of the title string.
        card = Record(title=u'no sync', description=u'Synced with Launchpad')
        self.assertTrue(should_sync_card(card, {'sync_cards': 'on'}))

    def test_should_sync_card_autosync_no(self):
        # When autosync is 'on', cards with no external card IDs
        # are not synced.
        card = Record(title=u'no sync', description=u'',
                      external_card_id=None)
        self.assertFalse(should_sync_card(card, {'sync_cards': 'on'}))

    def test_should_sync_card_autosync_yes(self):
        # When autosync is 'on', cards with external card IDs are synced.
        card = Record(title=u'no sync', description=u'',
                      external_card_id='11')
        self.assertTrue(should_sync_card(card, {'autosync': 'on',
                                                'sync_cards': 'on'}))

    def test_should_sync_card_autosync_nosync(self):
        # When autosync is 'on', cards with external card IDs are not
        # synced when they contain the no-sync marker.
        card = Record(title=u'(no-sync)', description=u'(no-sync)',
                      external_card_id='11')
        self.assertFalse(should_sync_card(card, {'autosync': 'on',
                                                 'sync_cards': 'on'}))

    def test_should_sync_card_no_sync_cards(self):
        # If the sync_cards config option is set to off,
        # should_sync_card() returns False without even looking at the
        # card.
        self.assertFalse(should_sync_card(None, {'sync_cards': 'off'}))

    def test_should_sync_card_no_sync_cards_default(self):
        # If the sync_cards config option isn't specified,
        # should_sync_card() returns False without even looking at the
        # card.
        self.assertFalse(should_sync_card(None, {}))

    def test_no_sync_lanes(self):
        # Cards in a 'no_sync' lane shouldn't get synced.
        board = FauxBoard()
        coding = board.addLane('coding')
        qa = board.addLane('qa')
        conf = {'no_sync_lanes': 'qa', 'autosync': 'on', 'sync_cards': 'on'}
        card1 = coding.addCard(FauxCard(external_card_id='42'))
        card2 = qa.addCard(FauxCard(external_card_id='84'))
        self.assertTrue(should_sync_card(card1, conf))
        self.assertFalse(should_sync_card(card2, conf))

    def test_no_sync_lanes_end(self):
        # Partial lane names can be specified in the 'no_sync_lanes'
        # config, so that every group's QA lane can be set to no_sync,
        # for example.
        board = FauxBoard()
        dev_qa = board.addLane('dev::qa')
        conf = {'no_sync_lanes': 'qa', 'sync_cards': 'on'}
        card = dev_qa.addCard(FauxCard(external_card_id='42'))
        self.assertFalse(should_sync_card(card, conf))

    def test_no_sync_lanes_multiple(self):
        # Multiple 'no_sync' lanes can be separated by comma in the
        # config.
        board = FauxBoard()
        qa = board.addLane('qa')
        review = board.addLane('review')
        conf = {'no_sync_lanes': 'qa,review', 'sync_cards': 'on'}
        card1 = qa.addCard(FauxCard(external_card_id='42'))
        card2 = review.addCard(FauxCard(external_card_id='84'))
        self.assertFalse(should_sync_card(card1, conf))
        self.assertFalse(should_sync_card(card2, conf))

    def test_no_sync_lanes_multiple_strip_whitespace(self):
        # Whitespace is ignored when specifying multiple lanes in the
        # 'no_sync_lanes' config.
        board = FauxBoard()
        qa = board.addLane('qa')
        review = board.addLane('review')
        conf = {'no_sync_lanes': ' qa ,  review ', 'sync_cards': 'on'}
        card1 = qa.addCard(FauxCard(external_card_id='42'))
        card2 = review.addCard(FauxCard(external_card_id='84'))
        self.assertFalse(should_sync_card(card1, conf))
        self.assertFalse(should_sync_card(card2, conf))

    def test_get_card_status_noop(self):
        # For a bug in 'New', 'Triaged', 'Confirmed' or 'Incomplete' statuses,
        # the status is unknown.
        self.assertEqual(None, get_card_status('New', [], None))
        self.assertEqual(None, get_card_status('Confirmed', [], None))
        self.assertEqual(None, get_card_status('Incomplete', [], None))
        self.assertEqual(None, get_card_status('Triaged', [], None))

    def test_get_card_status_coding_no_branch(self):
        # For a bug in 'In Progress' or 'Fix Committed' status, it is
        # considered to be in the coding state if there is no branch.
        branch_info = Record(status=None, target=None)
        self.assertEqual(
            CardStatus.CODING,
            get_card_status('In Progress', [], branch_info))
        self.assertEqual(
            CardStatus.CODING,
            get_card_status('Fix Committed', [], branch_info))

    def test_get_card_status_coding_branch(self):
        # For a bug in 'In Progress' or 'Fix Committed' status, it is
        # considered to be in the coding state if there is a branch
        # and it is 'In Progress'.
        branch_info = Record(status='In Progress', target=None)
        self.assertEqual(
            CardStatus.CODING,
            get_card_status('In Progress', [], branch_info))
        self.assertEqual(
            CardStatus.CODING,
            get_card_status('Fix Committed', [], branch_info))

    def test_get_card_status_review(self):
        # For a bug in 'In Progress' or 'Fix Committed' status, it is
        # considered to be in the review phase if there is a branch
        # proposed for review.
        branch_info = Record(status='In Review', target=None)
        self.assertEqual(
            CardStatus.REVIEW,
            get_card_status('In Progress', [], branch_info))
        self.assertEqual(
            CardStatus.REVIEW,
            get_card_status('Fix Committed', [], branch_info))

    def test_get_card_status_landing(self):
        # For a bug in 'In Progress' or 'Fix Committed' status, it is
        # considered to be in the landing phase if there is a branch
        # approved for landing.
        branch_info = Record(status='Approved', target=None)
        self.assertEqual(
            CardStatus.LANDING,
            get_card_status('In Progress', [], branch_info))
        self.assertEqual(
            CardStatus.LANDING,
            get_card_status('Fix Committed', [], branch_info))

    def test_get_card_status_qa(self):
        # For a bug in the 'Fix Committed' status, it is considered
        # to be in the QA phase if there is a branch that has been merged
        # and the bug has a 'qa-needstesting' tag.
        branch_info = Record(status='Merged', target=None)
        self.assertEqual(
            CardStatus.QA,
            get_card_status('Fix Committed', ['qa-needstesting'], branch_info))

        # If the bug is still 'In Progress', it is still in the landing phase.
        self.assertEqual(
            CardStatus.LANDING,
            get_card_status('In Progress', ['qa-needstesting'], branch_info))


    def test_get_card_status_deploy(self):
        # For a bug in the 'Fix Committed' status, it is
        # considered to be in the deploy phase if there is a branch
        # that has been merged and the bug has no 'qa-needstesting' tag,
        # and the branch target is 'devel'
        branch_info = Record(status='Merged', target='devel')
        self.assertEqual(
            CardStatus.DEPLOY,
            get_card_status('Fix Committed', [], branch_info))

        # If the bug is still 'In Progress', it is still in the landing phase.
        self.assertEqual(
            CardStatus.LANDING,
            get_card_status('In Progress', [], branch_info))

    def test_get_card_status_downtime_deploy(self):
        # For a bug in the 'Fix Committed' status, it is
        # considered to be in the downtime-deploy phase if there is a branch
        # that has been merged and the bug has no 'qa-needstesting' tag,
        # and the branch target is 'db-devel'
        branch_info = Record(status='Merged', target='db-devel')
        self.assertEqual(
            CardStatus.DOWNTIME_DEPLOY,
            get_card_status('Fix Committed', [], branch_info))

        # If the bug is still 'In Progress', it is still in the landing phase.
        self.assertEqual(
            CardStatus.LANDING,
            get_card_status('In Progress', [], branch_info))

    def test_get_card_status_done(self):
        # For a bug in 'Fix Released', 'Opinion', 'Won't fix', 'Invalid' or
        # or 'Expired' statuses, it is considered completely done.
        self.assertEqual(
            CardStatus.DONE, get_card_status('Fix Released', [], None))
        self.assertEqual(
            CardStatus.DONE, get_card_status('Opinion', [], None))
        self.assertEqual(
            CardStatus.DONE, get_card_status("Won't Fix", [], None))
        self.assertEqual(
            CardStatus.DONE, get_card_status('Invalid', [], None))
        self.assertEqual(
            CardStatus.DONE, get_card_status('Expired', [], None))


class FindLaneNextTest(unittest.TestCase):

    def test_find_card_target_lane_no_match(self):
        # Lane path needs to end with the string we are looking for.
        coding = FauxLane('Dev::Coding')
        review = FauxLane('Dev::Review')
        coding.addNextLane(review)

        self.assertEqual(None, find_card_target_lane_next(coding, 'Landing'))

    def test_find_card_target_lane_exists(self):
        coding = FauxLane('Dev::Coding')
        review = FauxLane('Dev::Review')
        coding.addNextLane(review)
        landing = FauxLane('Dev::Landing')
        review.addNextLane(landing)

        self.assertEqual(landing, find_card_target_lane_next(
            coding, '::Landing'))

    def test_find_card_target_lane_branching(self):
        # All the potential next lanes are searched through recursively
        # for the matching next lane.
        qa = FauxLane('QA::Done')
        land = FauxLane('Land::Quick')
        slow_land = FauxLane('Land::Slow')
        qa.addNextLane(land)
        qa.addNextLane(slow_land)
        rollout = FauxLane('Rollout')
        slow_land.addNextLane(rollout)

        self.assertEqual(rollout, find_card_target_lane_next(qa, 'Rollout'))


class GetNewBugsTest(unittest.TestCase):

    def test_empty_board_no_bugs(self):
        # If the board is empty, and there are no new bugs, no
        # bugs are returned.
        board = FauxBoard(cards=[])
        bugs = []
        self.assertEqual([], list(get_new_bugs(board, bugs)))

    def test_nonempty_board_no_bugs(self):
        # If the board has bugs, but there are no new bugs in the passed
        # in list, no bugs are returned.
        board = FauxBoard(
            cards=[FauxCard(), FauxCard(external_card_id="123")])
        bugs = []
        self.assertEqual([], list(get_new_bugs(board, bugs)))

    def test_no_new_bugs(self):
        # If all the passed in bugs are already in the board, no bugs
        # are returned.
        board = FauxBoard(
            cards=[FauxCard(), FauxCard(external_card_id="123")])
        bugs = [FauxBug(bug_id=123)]
        self.assertEqual([], list(get_new_bugs(board, bugs)))

    def test_new_bugs(self):
        # The passed in bugs that aren't in the board are returned.
        board = FauxBoard(
            cards=[FauxCard(), FauxCard(external_card_id="123")])
        bugs = [FauxBug(bug_id=123), FauxBug(bug_id=457)]
        self.assertEqual([bugs[1]], list(get_new_bugs(board, bugs)))


class FauxLaunchpadProject:

    def __init__(self, bug_tasks=None):
        if bug_tasks is None:
            bug_tasks = []
        self.bug_tasks = bug_tasks

    def searchTasks(self, tags=None):
        bug_tasks = self.bug_tasks
        if tags is not None:
            tags = set(tags)
            bug_tasks = [
                bug_task for bug_task in bug_tasks
                if tags.intersection(bug_task.bug.tags) == tags]
        return bug_tasks


class CreateNewCardsTest(unittest.TestCase):

    def test_no_bugs(self):
        # If no bugs are found, no cards are created.
        board = FauxBoard()
        bug_lane = board.addLane('bugs')
        project = FauxLaunchpadProject(bug_tasks=[])
        create_new_cards(board, project, 'kanban')
        self.assertEqual([], board.cards)

    def test_new_bugs_card_added(self):
        # If a new bug is found, a card is created for it, having the
        # bug id, title and description saved.
        board = FauxBoard()
        bug_lane = board.addLane('bugs')
        project = FauxLaunchpadProject(
            bug_tasks=[FauxBugTask(
                           bug_id=42, title='Test bug',
                           description='Test bug description.',
                           tags=['kanban'])])
        create_new_cards(board, project, 'kanban')
        [card] = board.cards
        self.assertEqual('Test bug', card.title)
        self.assertEqual('Test bug description.', card.description)
        self.assertEqual('42', card.external_card_id)
        self.assertEqual(board.default_cardtype.id, card.type_id)
        self.assertTrue(card.saved)

    def test_new_bugs_tag_removed(self):
        # After a card is created from a bug, the bug tag is removed, so
        # that having a lot of open bugs won't slow down the process.
        board = FauxBoard()
        bug_lane = board.addLane('bugs')
        project = FauxLaunchpadProject(
            bug_tasks=[FauxBugTask(
                           bug_id=42, title='Test bug',
                           description='Test bug description.',
                           tags=['kanban'])])
        create_new_cards(board, project, 'kanban')
        [bug_task] = project.bug_tasks
        self.assertEqual([], bug_task.bug.tags)
        self.assertTrue(bug_task.bug.saved)

    def test_new_bugs_tagged_only(self):
        # Cards are created only for bugs having the bug tag that is
        # passed in to create_new_cards().
        board = FauxBoard()
        bug_lane = board.addLane('bugs')
        project = FauxLaunchpadProject(
            bug_tasks=[FauxBugTask(
                           bug_id=21, title='Tagged with other tag',
                           tags=['sometag']),
                       FauxBugTask(
                           bug_id=42, title='Tagged with right tag',
                           tags=['kanban']),
                       FauxBugTask(
                           bug_id=42, title='Tagged with no tags',
                           tags=[])])
        create_new_cards(board, project, 'kanban')
        [card] = board.cards
        self.assertEqual('Tagged with right tag', card.title)
        self.assertEqual('42', card.external_card_id)

    def test_new_bugs_custom_card_type(self):
        # It's possible to specify which card type that the created card
        # should have, instead of the default one.
        board = FauxBoard()
        bug_lane = board.addLane('bugs')
        project = FauxLaunchpadProject(
            bug_tasks=[FauxBugTask(
                           bug_id=42, title='Test bug',
                           description='Test bug description.',
                           tags=['kanban'])])
        feature_type = FauxCardType(id=2, name='Feature', is_default=False)
        board.cardtypes[feature_type.id] = feature_type
        create_new_cards(board, project, 'kanban', cardtype_name='Feature')
        [card] = board.cards
        self.assertEqual(feature_type.id, card.type_id)
        self.assertTrue(card.saved)

    def test_new_bugs_custom_card_lane(self):
        # It's possible to specify to which lane the card should be
        # added.
        board = FauxBoard()
        bug_lane = board.addLane('bugs')
        bug_lane = board.addLane('incoming')
        project = FauxLaunchpadProject(
            bug_tasks=[FauxBugTask(
                           bug_id=42, title='Test bug',
                           description='Test bug description.',
                           tags=['kanban'])])
        feature_type = FauxCardType(id=2, name='Feature', is_default=False)
        board.cardtypes[feature_type.id] = feature_type
        create_new_cards(
            board, project, 'kanban', cardtype_name='Feature',
            cardlane_path='incoming')
        [card] = board.cards
        self.assertEqual('incoming', card.lane.path)


class GroupsConfigTest(unittest.TestCase):

    def test_parse_group_conf_empty_file(self):
        # If the group config file is empty, parse_groups_config()
        # returns and empty dict.
        sample_conf = ''

        config = parse_config_text(sample_conf)

        self.assertEqual({}, parse_groups_config(config))

    def test_parse_group_conf_no_members(self):
        # If the config file doesn't contain any members,
        # parse_groups_config() returns an empty dict.
        sample_conf = '''
            [Test group]
        '''

        config = parse_config_text(sample_conf)

        self.assertEqual({}, parse_groups_config(config))

    def test_parse_group_conf_single_member(self):
        # If the config file contains a member, parse_groups_config()
        # returns a dict mapping the member to the group.
        sample_conf = '''
            [Test group]
            members = foo
        '''

        config = parse_config_text(sample_conf)

        self.assertEqual({"foo": "Test group"}, parse_groups_config(config))

    def test_parse_group_conf_multiple_members(self):
        # Multiple members in a group are separated by comma.
        sample_conf = '''
            [Test group]
            members = foo,bar
        '''

        config = parse_config_text(sample_conf)

        self.assertEqual(
            {"foo": "Test group", "bar": "Test group"},
            parse_groups_config(config))

    def test_parse_group_conf_multiple_members_whitespace(self):
        # Whitespace is ignored when separating members with comma.
        sample_conf = '''
            [Test group]
            members =   foo ,  bar
        '''

        config = parse_config_text(sample_conf)

        self.assertEqual(
            {"foo": "Test group", "bar": "Test group"},
            parse_groups_config(config))

    def test_parse_group_conf_multiple_groups(self):
        # A groups config can contain multiple groups.
        sample_conf = '''
            [Test group 1]
            members = foo
            [Test group 2]
            members = bar
        '''

        config = parse_config_text(sample_conf)

        self.assertEqual(
            {"foo": "Test group 1", "bar": "Test group 2"},
            parse_groups_config(config))



class LaunchpadUsersForBoardTest(unittest.TestCase):

    def test_no_groups_config(self):
        # If the config doesn't contain an entry for a groups config,
        # ane empty dict is created for the group mapping.
        lp = None
        board = FauxBoard(users=[])
        config = {}
        lp_users = LaunchpadUsersForBoard()
        lp_users.set_up_users(lp, board, config)

        self.assertEqual({}, lp_users.lp_to_group)

    def test_with_groups_config(self):
        # If there is an entry for a groups config in the board config,
        # it is read and a groups mapping is created.
        groups_conf = '''
            [Group 1]
            members = foo,bar
            [Group 2]
            members = baz
        '''
        groups_file = tempfile.NamedTemporaryFile()
        groups_file.write(dedent(groups_conf))
        groups_file.flush()

        lp = None
        board = FauxBoard(users=[])
        config = {"groups_config_file": groups_file.name}
        lp_users = LaunchpadUsersForBoard()
        lp_users.set_up_users(lp, board, config)

        self.assertEqual(
            {"foo": "Group 1", "bar": "Group 1", "baz": "Group 2"},
            lp_users.lp_to_group)


class FauxPerson:


    def __init__(self, name):
        self.name = name


class MoveCardOmnidirectionalTest(unittest.TestCase):
    """Tests for move_card() with omnidirectional card moves turned on."""

    def setUp(self):
        self.bconf = {
            "omnidirectional_card_moves": "on",
            "coding_lanes": "coding",
            "review_lanes": "review",
            "landing_lanes": "landing",
            "qa_lanes": "qa",
            "deploy_lanes": "deploy",
            "downtime_deploy_lanes": "downtime",
            "done_lanes": "done",
        }
        self.board = FauxBoard()
        self.backlog = self.board.addLane("backlog")
        self.board.addLane("coding")
        self.review = self.board.addLane("review")
        self.board.addLane("landing")
        self.qa = self.board.addLane("qa")
        self.board.addLane("deploy")
        self.board.addLane("downtime")
        self.board.addLane("done")
        self.lp_users = LaunchpadUsersForBoard()

    def test_coding(self):
        # CODING cards are moved to the coding lane.
        card = self.backlog.addCard(FauxCard())
        move_card(card, CardStatus.CODING, self.bconf, [], self.lp_users)
        self.assertEqual("coding", card.moved_to.path)

    def test_review(self):
        # REVIEW cards are moved to the review lane.
        card = self.backlog.addCard(FauxCard())
        move_card(card, CardStatus.REVIEW, self.bconf, [], self.lp_users)
        self.assertEqual("review", card.moved_to.path)

    def test_landing(self):
        # LANDING cards are moved to the landing lane.
        card = self.backlog.addCard(FauxCard())
        move_card(card, CardStatus.LANDING, self.bconf, [], self.lp_users)
        self.assertEqual("landing", card.moved_to.path)

    def test_qa(self):
        # QA cards are moved to the qa lane.
        card = self.backlog.addCard(FauxCard())
        move_card(card, CardStatus.QA, self.bconf, [], self.lp_users)
        self.assertEqual("qa", card.moved_to.path)

    def test_deploy(self):
        # DEPLOY cards are moved to the deploy lane.
        card = self.backlog.addCard(FauxCard())
        move_card(card, CardStatus.DEPLOY, self.bconf, [], self.lp_users)
        self.assertEqual("deploy", card.moved_to.path)

    def test_downtime(self):
        # DOWNTIME_DEPLOY cards are moved to the downtime lane.
        card = self.backlog.addCard(FauxCard())
        move_card(
            card, CardStatus.DOWNTIME_DEPLOY, self.bconf, [], self.lp_users)
        self.assertEqual("downtime", card.moved_to.path)

    def test_done(self):
        # DONE cards are moved to the done lane.
        card = self.backlog.addCard(FauxCard())
        move_card(card, CardStatus.DONE, self.bconf, [], self.lp_users)
        self.assertEqual("done", card.moved_to.path)

    def test_backwards(self):
        # Cards can be moved backwards as well, e.g. go from QA to
        # CODING, if an issue is found.
        card = self.qa.addCard(FauxCard())
        move_card(card, CardStatus.CODING, self.bconf, [], self.lp_users)
        self.assertEqual("coding", card.moved_to.path)

    def test_group_no_assignee(self):
        # If a target lane requires a group, and the card doesn't have
        # an assignee, the card isn't moved.
        self.bconf["coding_lanes"] = "${group}::coding"
        self.board.addLane("test group::coding")
        card = self.backlog.addCard(FauxCard())
        move_card(card, CardStatus.CODING, self.bconf, [], self.lp_users)
        self.assertEqual(None, card.moved_to)

    def test_group_assignee_in_group(self):
        # If a target lane requires a group, and the card has an
        # assignee that is part of a group, the card is moved to the
        # group's lane.
        self.bconf["coding_lanes"] = "${group}::coding"
        self.board.addLane("test group::coding")
        card = self.backlog.addCard(FauxCard())
        self.lp_users.lp_to_group["foo"] = "test group"
        move_card(
            card, CardStatus.CODING, self.bconf, [FauxPerson("foo")],
            self.lp_users)
        self.assertEqual("test group::coding", card.moved_to.path)

    def test_group_assignee_not_in_group(self):
        # If a target lane requires a group, and the card has an
        # assignee that is not part of a group, the card isn't moved.
        self.bconf["coding_lanes"] = "${group}::coding"
        self.board.addLane("test group::coding")
        card = self.backlog.addCard(FauxCard())
        self.lp_users.lp_to_group["foo"] = "test group"
        move_card(
            card, CardStatus.CODING, self.bconf, [FauxPerson("bar")],
            self.lp_users)
        self.assertEqual(None, card.moved_to)

    def test_no_move_lanes(self):
        # Cards in a "no_move" lane doesn't get moved by move_card().
        self.bconf["no_move_lanes"] = "qa"
        card = self.qa.addCard(FauxCard())
        move_card(card, CardStatus.CODING, self.bconf, [], self.lp_users)
        self.assertEqual(None, card.moved_to)

    def test_no_move_lanes_end(self):
        # Partial lane names can be specified in the "no_move_lanes"
        # config, so that every group's QA lane can be set to no_move,
        # for example.
        dev_qa = self.board.addLane("dev::qa")
        self.bconf["no_move_lanes"] = "qa"
        card = dev_qa.addCard(FauxCard())
        move_card(card, CardStatus.CODING, self.bconf, [], self.lp_users)
        self.assertEqual(None, card.moved_to)

    def test_no_move_lanes_multiple(self):
        # Multiple "no_move" lanes can be separated by comma in the
        # config.
        self.bconf["no_move_lanes"] = "qa,review"
        card1 = self.qa.addCard(FauxCard())
        card2 = self.review.addCard(FauxCard())
        move_card(card1, CardStatus.CODING, self.bconf, [], self.lp_users)
        move_card(card2, CardStatus.CODING, self.bconf, [], self.lp_users)
        self.assertEqual(None, card1.moved_to)
        self.assertEqual(None, card2.moved_to)

    def test_no_move_lanes_multiple_strip_whitespace(self):
        # Whitespace is ignored when specifying multiple lanes in the
        # "no_move_lanes" config.
        self.bconf["no_move_lanes"] = " qa ,  review "
        card1 = self.qa.addCard(FauxCard())
        card2 = self.review.addCard(FauxCard())
        move_card(card1, CardStatus.CODING, self.bconf, [], self.lp_users)
        move_card(card2, CardStatus.CODING, self.bconf, [], self.lp_users)
        self.assertEqual(None, card1.moved_to)
        self.assertEqual(None, card2.moved_to)


class MoveCardTest(unittest.TestCase):

    def test_should_move_not_if_disabled(self):
        board = FauxBoard(cards=[], is_archived=False)
        lane = FauxLane(board=board)
        card = FauxCard(lane=lane)
        board.cards.append(card)
        self.assertFalse(should_move_card(card, {'move_cards': 'off'}))

    def test_should_move_not_if_disabled_and_the_default_is_disabled(self):
        board = FauxBoard(cards=[], is_archived=False)
        lane = FauxLane(board=board)
        card = FauxCard(lane=lane)
        board.cards.append(card)
        self.assertFalse(should_move_card(card, {}))

    def test_should_move_not_if_archived(self):
        board = FauxBoard(cards=[], is_archived=True)
        lane = FauxLane(board=board)
        card = FauxCard(lane=lane)
        board.cards.append(card)
        self.assertFalse(should_move_card(card, {'move_cards': 'on'}))

    def test_should_move_if_enabled_and_not_archived(self):
        board = FauxBoard(cards=[], is_archived=False)
        lane = FauxLane(board=board)
        card = FauxCard(lane=lane)
        board.cards.append(card)
        self.assertTrue(should_move_card(card, {'move_cards': 'on'}))
